//: blocco_jacobi_par.h Classe ereditata da blocco_jacobi per il calcolo parallelo.
//. La classe PBloccoJacobi rappresenta una singola unita' matriciale
//. adatta al calcolo dell'algoritmo di Jacobi su una architettura parallela.
//. La classe e' astratta e serve da interfaccia per implementazioni specifiche.
#ifndef _BLOCCO_JACOBI_PAR
#define _BLOCCO_JACOBI_PAR

#include <stdlib.h>
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include "blocco_jacobi.h"
#include "mpi.h"

class PBloccoJacobi : public BloccoJacobi {
	protected:
	char host_name[MPI_MAX_PROCESSOR_NAME];	// nome dell'host a cui appartiene l'oggetto.
	int nameLength;	// lunghezza del nome dell'host.
	int id;	// rank del processo a cui appartiene l'oggetto.
	int nProcesses;	// numero totale dei processi nel gruppo.
	
	unsigned int passo_sincronia;	// specifica ogni quante iterazioni i processi
									// si sincronizzano aggiornando i bordi delle
									// matrici.
	unsigned int num_sincronie;	// indica il numero di sincronie effettuate dai
								// processi.
	unsigned int num_iterazioni;	// indica il numero di iterazioni effettuate dai
									// processi.
	float tempo_calcolo;		// indica il tempo (secondi) impiegato per il
								// calcolo di Jacobi,
	float tempo_setup;		// indica il tempo (secondi) impiegato per eseguire
							// il protocollo di riconoscimento dei confini.
	float errore_locale;		// indica l'errore relativo raggiunto dall'oggetto.
	float errore_globale;		// indica l'errore relativo massimo rispetto a tutti
								// i processi del gruppo.
	
	public:
	PBloccoJacobi(int ux, int uy, int dx, int dy) :
		BloccoJacobi(ux, uy, dx, dy) {

		passo_sincronia = 1;
		num_iterazioni = 0;
		num_sincronie = 0;
		tempo_calcolo = 0.0;
		tempo_setup = 0.0;

		MPI_Comm_size(MPI_COMM_WORLD, &nProcesses);
		MPI_Comm_rank(MPI_COMM_WORLD, &id);
		MPI_Get_processor_name(host_name, &nameLength);
	}
	
	virtual unsigned int calcola_jacobi() {
	}
	
	virtual void ai_limiti(float nord, float sud, float est, float ovest) {}
	
	void set_passo_sincronia(unsigned int n) { passo_sincronia = n; }
	
	unsigned int get_num_sincronie() { return num_sincronie; }
	float get_tempo_calcolo() { return tempo_calcolo; }
	float get_tempo_setup() { return tempo_setup; }
	float get_tempo_totale() { return tempo_calcolo + tempo_setup; }
	float get_errore() { return errore_globale; }
	
	virtual void mostra_risultato() {};
	
	void print_info() {
		unsigned int c=0;
		cout << "Matrice M(" << righe << "," << colonne << ")" 
		     << " da (" << ux << "," << uy << ") a (" << dx << ","
		     << dy << ") sulla macchina " << host_name << " processo #"
		     << id << endl;
		     
		BloccoJacobi::mostra_risultato();
	}
};

#endif _BLOCCO_JACOBI_PAR