//: blocco_jacobi_par_striscia.h Classe ereditata da PBloccoJacobi.
//. Implementa l'interfaccia di PBloccoJacobi per il calcolo parallelo
//. di Jacobi suddividendo la matrice originaria in strisce.
//. Implementa i protocolli di riconoscimento delle strisce e di recupero
//. dei dati.
#ifndef _BLOCCO_JACOBI_PAR_STRISCIA
#define _BLOCCO_JACOBI_PAR_STRISCIA

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include "mpi.h"
#include "blocco_jacobi_par.h"

using namespace std;	// necessario per usare oggetti ofstream.

class PBloccoJacobiStriscia : public PBloccoJacobi {

	protected:
	ofstream out;	// stream su cui confluiscono le informazioni di debug.
	int N;	// lato della matrice e quindi altezza delle strisce.
	float *sx_col, *dx_col;	// vettori contenenti i confini "stranieri", rispettivamente
							// il sinitro e il destro.
	unsigned int min_x, max_x;	// delimitano le strisce su cui avverra' il calcolo,
								// vengono quindi tolti i limiti della matrice.
	unsigned int start_x, end_x; // start_x ed end_x sono le coordinate
		                     // entro cui posso calcolare jacobi
		                     // senza richiedere dati ad altri
		                     // processi
	unsigned int next_sincronia;	// indica il numero di iterazione in cui
									// avverra' la prossima sincronia tra i processi.
	
	int proc_start_x; // id del processo che possiede il confine sinistro
	int proc_end_x; // id del processo che possiede il confine destro
	
	unsigned int clienti; //numero di processi a cui dovro' spedire delle strisce
	
	//: setup_confini() Implementa il protocollo di riconoscimento dei confini.
	//. Il metodo assegnera' i valori appropriati agli attributi proc_start_x,
	//. proc_end_x, clienti, max_x, min_x, start_x e end_x.
	void setup_confini() {
		clienti = 0;

		tempo_setup = MPI_Wtime();
		
		if (dx == N) 
			max_x = dx-1;
		else max_x = dx;
		
		if (ux == 1)
			min_x = ux+1;
		else min_x = ux;
		
		start_x = min_x;
		end_x = max_x;
		
		if (id!=0) {
			// Spedisco quello che ho (i due interi ux e dx) a root.
			MPI_Send(&ux, 1, MPI_INT, 0, 1, MPI_COMM_WORLD);
			MPI_Send(&dx, 1, MPI_INT, 0, 1, MPI_COMM_WORLD);
			#ifdef DEBUG
			out << "Spedite informazioni sui confini #" << id << endl;
			out.flush();
			#endif
			
			// Spedisco quello che mi serve (ux-1, dx+1, -1 se non mi serve)
			// a root.
			{
				int temp;
				
				if (ux != 1) {
					temp = ux-1;
					start_x = min_x+1;
				} else temp = -1;
				MPI_Send(&temp, 1, MPI_INT, 0, 2, MPI_COMM_WORLD);

				if (dx != N) {
					temp = dx+1;
					end_x = max_x-1;
				} else temp = -1;
				MPI_Send(&temp, 1, MPI_INT, 0, 2, MPI_COMM_WORLD);
				
				#ifdef DEBUG
				out << "Spedite informazioni sui confini di cui ho bisogno #" 
				     << id << endl;
				out.flush();
				#endif
			}
			
			// Ricevo da root gli id dei processi che hanno i confini
			// che mi servono
			{
				MPI_Status status_start_x;
				MPI_Status status_end_x;
				if (min_x != start_x)
					MPI_Recv(&proc_start_x, 1, MPI_INT, 0, 3, MPI_COMM_WORLD,
						&status_start_x);
				if (max_x != end_x)
					MPI_Recv(&proc_end_x, 1, MPI_INT, 0, 3, MPI_COMM_WORLD,
						&status_end_x);
			}
			#ifdef DEBUG
			out << "Ricevute informazioni sui possessori dei confini " 
			     << proc_start_x << "," << proc_end_x << endl;
			out.flush();
			#endif
			
			// Ricevo da root il numero dei miei confini che dovro'
			// spedire agli altri processi.
			{
				MPI_Status status_clienti;
				MPI_Recv(&clienti, 1, MPI_INT, 0, 4, MPI_COMM_WORLD,
					&status_clienti);
					
				#ifdef DEBUG
				out << "Dovro' spedire " << clienti << " confini #"
				     << id << endl;
				out.flush();
				#endif
			}			
		} else { 
			// Creo e gestisco le liste di proprietari-richieste
			int *proprietari; // proprietario[i] = id di
			                  // chi possiede il confine i.
			int *richiesti; // richiesti[i] = id di
			                // chi ha bisgono del confine i.
			int *qta; // qta[i] = numero di confini che il processo i
			          // dovra' spedire agli altri processi.
			int temp;
			MPI_Status status;
			
			qta = new int[nProcesses];
			richiesti = new int[N+1];
			proprietari = new int[N+1];
			for (unsigned int i=0; i<nProcesses; i++)
				qta[i] = 0;
			
			for (unsigned int i=1; i<=N; i++) {
				proprietari[i] = -1;
				richiesti[i] = -1;
			}
			
			// inserisco i miei dati
			proprietari[ux] = 0;
			proprietari[dx] = 0;
			if (ux!=1) {
				richiesti[ux-1] = 0;
				start_x = min_x+1;
			}
			if (dx!=N) {
				richiesti[dx+1] = 0;
				end_x = max_x-1;
			}
		
			// Recupero i dati sui confini posseduti dai processi
			for (unsigned int i=1; i<nProcesses; i++) {
				MPI_Recv(&temp, 1, MPI_INT, i, 1, MPI_COMM_WORLD,
				&status);
				proprietari[temp] = i;
				#ifdef DEBUG
				out << "Confine " << temp << " appartiene a "
				     << i << endl;
				out.flush();
				#endif
				MPI_Recv(&temp, 1, MPI_INT, i, 1, MPI_COMM_WORLD,
				&status);
				proprietari[temp] = i;
				#ifdef DEBUG
				out << "Confine " << temp << " appartiene a "
				     << i << endl;
				out.flush();
				#endif
			}
			
			// Recupero i dati sui confini richiesti dai processi
			for (unsigned int i=1; i<nProcesses; i++) {
				MPI_Recv(&temp, 1, MPI_INT, i, 2,
				MPI_COMM_WORLD, &status);
				if (temp != -1) {
					richiesti[temp] = i;
					#ifdef DEBUG
					out << "Confine " << temp << " richiesto da "
					     << i << endl;
					out.flush();
					#endif
				}
				MPI_Recv(&temp, 1, MPI_INT, i, 2,
				MPI_COMM_WORLD, &status);
				if (temp != -1) {
					richiesti[temp] = i;
					#ifdef DEBUG
					out << "Confine " << temp << " richiesto da "
					     << i << endl;
					out.flush();
					#endif
				}
			}
			
			#ifdef DEBUG
			out << "proprietari: ";
			for (unsigned int i=1; i<=N; i++)
				out << proprietari[i] << ", ";
			out << endl << "richiesti: ";
			for (unsigned int i=1; i<=N; i++)
				out << richiesti[i] << ", ";
			out << endl;
			out.flush();
			#endif
			
			// Spedisco ai vari processi gli id dei processi
			// che possiedono i confini da loro richiesti
			// controllando prima (nei due while) le richieste di root.
			// (sarebbe inutile spedirsi un messaggio).
			{
				unsigned int i=1;
				while (richiesti[i] != 0 && i<=N) i++;
				if (richiesti[i] == 0) {
					if (start_x != min_x)
						proc_start_x = proprietari[i];
					else
						proc_end_x = proprietari[i];
					richiesti[i] = -1;
					qta[proprietari[i]]++;
					clienti++;
					
					#ifdef DEBUG
					out << "Confine " << i << " posseduto da "
						<< proprietari[i] << endl;
					out.flush();
					#endif
				}
				while (richiesti[i] != 0 && i<=N) i++;
				if (richiesti[i] == 0) {
					proc_end_x = proprietari[i];
					richiesti[i] = -1;
					qta[proprietari[i]]++;
					clienti++;
					
					#ifdef DEBUG
					out << "Confine " << i << " posseduto da "
						<< proprietari[i] << endl;
					out.flush();
					#endif
				}
				
				#ifdef DEBUG
				out << "Numero miei clienti " << clienti << endl;
				out.flush();
				#endif
			}
			
			int own, req;
			for (unsigned int i=1; i<=N; i++) {
				if ((own=proprietari[i]) != -1)
					if ((req=richiesti[i]) != -1) {
						MPI_Send(&own, 1, MPI_INT, req,
					  	3, MPI_COMM_WORLD);
					  	
					  	qta[own]++;
					  	
						#ifdef DEBUG
						out << "Spedito id " << own
						     << " al processo " << req
						     << endl;
						out.flush();
						#endif
					}
			}
			
			// Ora spedisco ad ogni processo la quantita' di confini che
			// dovra' spedire ad altri processi.
			for (unsigned int i=1; i<nProcesses; i++)
				MPI_Send(qta+i, 1, MPI_INT, i, 4, MPI_COMM_WORLD);
			
			delete []qta;
			delete []richiesti;
			delete []proprietari;

			tempo_setup = MPI_Wtime() - tempo_setup;
		}
		
		#ifdef DEBUG
		out << "Finita raccolta informazioni iniziali #" << id << endl;
		out.flush();
		#endif
	}

	public:

	PBloccoJacobiStriscia(int ux, int uy, int dx, int dy) :
		PBloccoJacobi(ux, uy, dx, dy), N(uy) {
		
		if (dx == N) 
			max_x = dx-1;
		else max_x = dx;
		
		if (ux == 1)
			min_x = ux+1;
		else min_x = ux;
		
		proc_start_x = id;
		proc_end_x = id;

		sx_col = dx_col = 0;

		char filename[100];
		sprintf(filename, "trace%d.log", id);

		out.open(filename, ofstream::out);
		out << "# Processo " << id << " sulla macchina " << host_name << endl;
	}

	~PBloccoJacobiStriscia() {
		out.close();
	}
	
	void ai_limiti(float nord, float sud, float est, float ovest) {
		set_all(0.0);
		
		if (ux == 1) {
			for (unsigned int i=1; i<=righe; i++) {
				it_att[i][1] = est;
				it_new[i][1] = est;
			}
		}
		
		if (dx == N) {
			for (unsigned int i=1; i<=righe; i++) {
				it_att[i][colonne] = ovest;
				it_new[i][colonne] = ovest;
			}
		}
		
		if (uy == N) {
			for (unsigned int i=1; i<=colonne; i++) {
				it_att[1][i] = nord;
				it_new[1][i] = nord;
			}
		}
		
		if (dy == 1) {
			for (unsigned int i=1; i<=colonne; i++) {
				it_att[righe][i] = sud;
				it_new[righe][i] = sud;
			}
		}
	}

	//: step_jacobi() Esegue una iterazione dell'algoritmo di Jacobi.
	//. Aggiorna le matrici secondo l'algoritmo di Jacobi, richiedendo
	//. eventuali confini stranieri necessari al calcolo ai rispettivi
	//. proprietari e soddisfando le richieste altrui.
	void step_jacobi() {

		int temp_min_x, temp_max_x;	// variabili di appoggio
		MPI_Request req_min_x, req_max_x, rcv_min_x, rcv_max_x;	// strutture dati
																// necessarie alla
																// comunicazione.
		MPI_Request *Mreq_clienti = 0;
		int *req_clienti = 0;
		
		int num_req = 0; // numero di richieste che faccio
		                 // a processi esterni;

		num_iterazioni++;

		#ifdef DEBUG
		out << endl << "Iterazione numero: " << num_iterazioni
			<< "\t Prossima sincronia: " << next_sincronia << endl
			<< "**** Situazione variabili ****" << endl
			<< "min_x: " << min_x << "\tmax_x: " << max_x << endl
			<< "start_x: " << start_x << "\tend_x:" << end_x <<
			endl;
		out.flush();
		#endif
			
		if (num_iterazioni == next_sincronia) {	// avvio una nuova procedura di sincronia.
			
			num_sincronie++;
		
			if (sx_col) delete []sx_col;	// cancello i vecchi confini
			if (dx_col) delete []dx_col;
			
			// Richiedo con chiamata asincrona i confini di cui ho bisogno
			// e ne inizio l'attesa asincrona.
			temp_min_x = temp_max_x = -1;
			sx_col = dx_col = 0;
			if (start_x != min_x) {
				temp_min_x = ux-1;
				num_req++;
				MPI_Isend(&temp_min_x, 1, MPI_INT, proc_start_x, 10, 
					MPI_COMM_WORLD, &req_min_x);
				sx_col = new float[N];
				MPI_Irecv(sx_col, N, MPI_FLOAT, proc_start_x, 20,
					MPI_COMM_WORLD, &rcv_min_x);
			}
			if (end_x != max_x) {
				temp_max_x = dx+1;
				num_req++;
				MPI_Isend(&temp_max_x, 1, MPI_INT, proc_end_x, 10,
					MPI_COMM_WORLD, &req_max_x);
				dx_col = new float[N];
				MPI_Irecv(dx_col, N, MPI_FLOAT, proc_end_x, 20,
					MPI_COMM_WORLD, &rcv_max_x);
			}
			#ifdef DEBUG
			out << "#" << id << " richiede confini: " << temp_min_x <<
				"," << temp_max_x << " a " << proc_start_x <<
				"," << proc_end_x << endl;
			out.flush();
			#endif
		
			// Inizio una ricezione asincrona sulle richieste altrui.
			Mreq_clienti = new MPI_Request[clienti];
			req_clienti = new int[clienti];
			for (unsigned int i=0; i<clienti; i++) {
				MPI_Irecv(req_clienti+i, 1, MPI_INT, MPI_ANY_SOURCE,
					10, MPI_COMM_WORLD, Mreq_clienti+i);
				#ifdef DEBUG
				out << "Iniziata attesa " << i << "-esimo cliente "
				     << "#" << id << endl;
				out.flush();
				#endif
			}
		}
				
		// Eseguo la parte di calcolo che non richiede dati
		// esterni al processo.
		for (unsigned int i=2; i<N; i++)
			for (unsigned int j=(start_x-ux+1);
			j<=(end_x-ux+1); j++) {
				it_new[i][j] = 0.25*(it_att[i-1][j] + 
				it_att[i][j+1] + it_att[i+1][j] + 
				it_att[i][j-1]);
			}
			
		if (num_iterazioni == next_sincronia) {	// completo la sincronia.
		
			// Attendo e servo le richieste altrui.
			for (unsigned int i=0; i<clienti; i++) {
				MPI_Status stat;
				MPI_Wait(Mreq_clienti+i, &stat);
				float *temp = new float[N];
				for (unsigned int j=1; j<=N; j++)
					temp[j] = it_att[j][req_clienti[i]-ux+1];
				MPI_Send(temp, N, MPI_FLOAT, stat.MPI_SOURCE, 20, 
					MPI_COMM_WORLD);
				delete []temp;
				
				#ifdef DEBUG
				out << "#" << id << " --" <<
					req_clienti[i]-ux+1 << "--> " << stat.MPI_SOURCE
				     << endl;
				out.flush();
				#endif
			}
		
			// Attendo i dati che ho richiesto precedentemente.
			if (sx_col) {
				MPI_Status stat;
				MPI_Wait(&rcv_min_x, &stat);
					#ifdef DEBUG
				out << "#" << id << " ricevuti dati da "
				     << stat.MPI_SOURCE << endl;
				out.flush();
				#endif
			}
		
			if (dx_col) {
				MPI_Status stat;
				MPI_Wait(&rcv_max_x, &stat);

				#ifdef DEBUG
				out << "#" << id << " ricevuti dati da "
				     << stat.MPI_SOURCE << endl;
				out.flush();
				#endif
			}
			
			if (Mreq_clienti) delete []Mreq_clienti;
			if (req_clienti) delete []req_clienti;

			next_sincronia += passo_sincronia;
		}

		// Eseguo i calcoli che coinvolgono l'eventuale confine
		// sinistro "straniero"
		if (start_x != min_x) {
			unsigned int col = (min_x-ux+1);
			for (unsigned int j=2; j<N; j++)
				it_new[j][col] = 0.25*(sx_col[j] +
				it_att[j+1][col] + it_att[j][col+1] +
				it_att[j-1][col]);
		}

		// Eseguo i calcoli che coinvolgono l'eventuale confine
		// destro "straniero"
		if (end_x != max_x) {
			unsigned int col = (max_x-ux+1);
			for (unsigned int j=2; j<N; j++)
				it_new[j][col] = 0.25*(dx_col[j] +
				it_att[j+1][col] + it_att[j][col-1] +
				it_att[j-1][col]);
		}
			
		// Calcolo l'errore relativo locale
		errore_locale = 0.0;
		for (unsigned int i=1; i<=righe; i++)
			for (unsigned int j=1 ;
			j<=colonne; j++) {
				errore_locale = max(errore_locale, fabs(it_new[i][j] - it_att[i][j]));
				it_att[i][j] = it_new[i][j];
			}

		#ifdef DEBUG
		out << "Errore relativo locale: " << errore_locale << endl;
		out.flush();
		#endif

		// Calcolo dell'errore relativo a livello globale.
		errore_globale = 0.0;
		MPI_Allreduce(&errore_locale, &errore_globale, 1,
			MPI_FLOAT, MPI_MAX, MPI_COMM_WORLD);
		#ifdef DEBUG
		out << "Errore relativo globale: " << errore_globale << endl;
		out.flush();
		#endif
	}

	//: calcola_jacobi() Esegue tante iterazioni di Jacobi fino ad arrivare all'errore
	//. desiderato.
	unsigned int calcola_jacobi() {

		setup_confini();

		MPI_Barrier(MPI_COMM_WORLD);

		tempo_calcolo = MPI_Wtime();
		num_sincronie = 0;
		next_sincronia = 1;
		num_iterazioni = 0;
		
		do {

			step_jacobi();
			MPI_Barrier(MPI_COMM_WORLD);
	
		} while(errore_globale > convergenza);
		
		
		tempo_calcolo = MPI_Wtime() - tempo_calcolo;
		return num_iterazioni;
	}
	
	//: mostra_risultato() implementa il protocollo di raccolta dei dati.
	//. Stampa sullo standard output la matrice totale.
	void mostra_risultato() {
	
		if (id != 0) {
			// Spedisco a root gli indici dei miei confini...
			MPI_Send(&ux, 1, MPI_INT, 0, 40, MPI_COMM_WORLD);
			MPI_Send(&dx, 1, MPI_INT, 0, 41, MPI_COMM_WORLD);
			
			// ...creo un vettore che li conterra'...
			float *temp = new float[righe*colonne];
			#ifdef DEBUG
			out << "Creato vettore temp[" << righe*colonne << "] #"
			     << id << endl;
			out.flush();
			#endif
			
			// ...lo riempio e...
			unsigned int c=0;
			for (unsigned int i=1; i<=righe; i++)
				for (unsigned int j=1; j<=colonne; j++)
					temp[c++] = it_att[i][j];
			#ifdef DEBUG
			out << "Riempito il vettore temp #" << id << endl;
			out.flush();
			#endif
			
			// ...lo spedisco a root.
			MPI_Send(temp, righe*colonne, MPI_FLOAT, 0, 42,
				MPI_COMM_WORLD);
			
			#ifdef DEBUG
			out << "Spediti " << righe*colonne << " dati a root"
			     << " #" << id << endl;
			out.flush();
			#endif
			delete []temp;
		} else {
		
			int osx, odx;	// indici dei confini delle matrici straniere.
			MPI_Status statsx, statdx, stattemp;
			float **mat;	// punta alla matrice totale.
			
			// creo la matrice totale.
			mat = new float*[N+1];
			for (unsigned int i=1; i<=N; i++)
				mat[i] = new float[N+1];
				
			#ifdef DEBUG
			out << "Creata matrice " << N << "x" << N << endl;
			out.flush();
			#endif
			
			// Ci inserisco i risultati delle mie computazioni
			unsigned int r,c;
			r=c=1;
			for (unsigned int i=dy; i<=uy; i++) {
				for (unsigned int j=ux; j<=dx; j++) {
					mat[i][j] = it_att[r][c];
					c++;
				}
				r++;
				c=1;
			}
			
			#ifdef DEBUG
			out << "Inseriti i miei dati (root)" << endl;
			out.flush();
			#endif
			
			// inserisco i dati altrui.
			for (unsigned int i=1; i<nProcesses; i++) {
				MPI_Recv(&osx, 1, MPI_INT, i, 40, MPI_COMM_WORLD,
					&statsx);
				MPI_Recv(&odx, 1, MPI_INT, i, 41, MPI_COMM_WORLD,
					&statdx);
				#ifdef DEBUG
				out << "Il processo " << statsx.MPI_SOURCE
				     << " possiede i confini " << osx << ","
				     << odx << endl;
				out.flush();
				#endif
				
				float *temp = new float[N*(odx-osx+1)];
				MPI_Recv(temp, N*(odx-osx+1), MPI_FLOAT, i, 42,
					MPI_COMM_WORLD, &stattemp);
				unsigned int k=0;
				for (unsigned int c=1; c<=N; c++)
					for (unsigned int j=osx; j<=odx; j++)
						mat[c][j] = temp[k++];
				delete []temp;
				
				#ifdef DEBUG
				out << "Ricevuti dati da " << stattemp.MPI_SOURCE
				     << endl;
				out.flush();
				#endif
			}

			// stampo la matrice sullo standard output.			
			for (unsigned int i=1; i<=N; i++) {
				for (unsigned int j=1; j<=N; j++)
					cout << mat[i][j] << "\t";

				delete []mat[i];
				cout << endl;

			}
			
			// cancello la matrice per liberare memoria.
			delete []mat;
		}
	}
};

#endif _BLOCCO_JACOBI_PAR_STRISCIA
