#include <iostream.h>
#include "mpi.h"
#include "blocco_jacobi_par.h"
#include "blocco_jacobi_par_striscia_din.h"
#include "datijacobi.h"

int main( int argc, char *argv[] ) {
	int myid, nProcesses, nameLength;
	unsigned int dinamico = 0;
	char process_name;

	// Inizializzo MPI.
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &nProcesses);
	MPI_Comm_rank(MPI_COMM_WORLD, &myid);

	unsigned int N = 1 / DELTA;
	unsigned int dim_strisce = N / nProcesses;
	unsigned int passo = 1;
	float convergenza = CONVERGENZA;

	if (argc < 5) {
		if (myid == 0) {
			cout << "Errore" << endl
				<< "Usare 'parallelo <N> <Convergenza> <Passo sincronia> d|s'"
				<< endl;
		}
		MPI_Finalize();
		exit(0);
	}

	// Conversione degli argomenti (N, Convergenza, passo di sincronia)
	N = atoi(argv[1]);
	convergenza = atof(argv[2]);
	passo = atoi(argv[3]);
	if (argv[4][0] == 'd') dinamico = 1;

	dim_strisce = N / nProcesses;

	if (myid == 0) {
		cout << endl << "# Inizio il calcolo con i seguenti parametri d'avvio"
		  << endl << "# N: " << N << "\t Convergenza: " << convergenza
		  << "\tPasso: " << passo << "\tNumero processi: " << nProcesses
		  << endl;
		if (dinamico) cout << "# Algoritmo dinamico" << endl;
		else cout << "# Algoritmo statico" << endl;
	}

	PBloccoJacobi *b;
	if (myid == 0)
		if (dinamico)
			b = new PDBloccoJacobiStriscia(1, N, dim_strisce+N%nProcesses,1);
		else
			b = new PBloccoJacobiStriscia(1, N, dim_strisce+N%nProcesses,1);
	else
		if (dinamico)
			b = new PDBloccoJacobiStriscia( (myid*dim_strisce+1)+(N%nProcesses),
				N, (myid*dim_strisce+1)+(N%nProcesses)+dim_strisce-1, 1);
		else
			b = new PBloccoJacobiStriscia( (myid*dim_strisce+1)+(N%nProcesses),
				N, (myid*dim_strisce+1)+(N%nProcesses)+dim_strisce-1, 1);

	b->ai_limiti(NORD,SUD,EST,OVEST);
	b->set_passo_sincronia(passo);
	b->set_convergenza(convergenza);
	unsigned int n = b->calcola_jacobi();

	if (myid==0) {
		cout << "# Numero iterazioni: " << n << endl;
		cout << "# Numero di sincronizzazioni: " << b->get_num_sincronie()
			<< endl;
		cout << "# Tempo impiegato per il calcolo: " << b->get_tempo_calcolo()
			<< " secondi" << endl;
		cout << "# Tempo impiegato per il setup: " << b->get_tempo_setup()
			<< " secondi" << endl;
		cout << "# Tempo totale: " << b->get_tempo_totale() << " secondi" << endl;
		cout << "# Errore relativo: " << b->get_errore() << endl;
	}
	
	b->mostra_risultato();
	
	delete b;

	MPI_Finalize();
}
